"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.AIClient = void 0;
class ClientPool {
    constructor(clients) {
        this.clients = clients;
        this.busyClients = new Set();
        this.queue = [];
    }
    getNextAvailableClient() {
        for (let client of this.clients) {
            if (!this.busyClients.has(client)) {
                return client;
            }
        }
        return null;
    }
    processQueue() {
        if (this.queue.length > 0) {
            const availableClient = this.getNextAvailableClient();
            if (availableClient) {
                const requestFunction = this.queue.shift();
                if (requestFunction) {
                    this.busyClients.add(availableClient);
                    requestFunction(availableClient).finally(() => {
                        this.busyClients.delete(availableClient);
                        this.processQueue();
                    });
                }
            }
        }
    }
    execute(request) {
        return __awaiter(this, void 0, void 0, function* () {
            const availableClient = this.getNextAvailableClient();
            if (availableClient) {
                this.busyClients.add(availableClient);
                availableClient;
                return yield request(availableClient)
                    .finally(() => {
                    this.busyClients.delete(availableClient);
                    this.processQueue();
                });
            }
            else {
                return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                    this.queue.push((client) => __awaiter(this, void 0, void 0, function* () {
                        try {
                            const result = request(client);
                            resolve(result);
                        }
                        catch (e) {
                            reject(e);
                        }
                    }));
                }));
            }
        });
    }
}
class AIClient {
    constructor(embeddingProviders, moderationProviders, enrichmentProviders, logger) {
        this.logger = logger;
        // Initialize pool maps
        this.embeddingPoolMap = this.initializePoolMap(embeddingProviders);
        this.moderationPoolMap = this.initializePoolMap(moderationProviders);
        this.enrichmentPoolMap = this.initializePoolMap(enrichmentProviders);
    }
    initializePoolMap(providers) {
        const poolMap = new Map();
        providers.forEach(provider => {
            const pool = new ClientPool(provider.clients);
            provider.models.forEach(model => {
                poolMap.set(model, pool); // Associate all models with the same pool
            });
        });
        return poolMap;
    }
    performEmbedding(createEmbeddingRequest, options, logMeta) {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            (_a = this.logger) === null || _a === void 0 ? void 0 : _a.debug("performEmbedding", logMeta);
            const model = createEmbeddingRequest.model;
            const pool = this.embeddingPoolMap.get(model); // Get the pool associated with the model
            if (!pool) {
                (_b = this.logger) === null || _b === void 0 ? void 0 : _b.error(`No client pool available for model ${model}`, {}, logMeta);
                throw new Error(`No client pool available for model ${model}`);
            }
            return yield pool.execute((client) => __awaiter(this, void 0, void 0, function* () {
                return yield client.performEmbedding(createEmbeddingRequest, options);
            }));
        });
    }
    performModeration(body, options, logMeta) {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            (_a = this.logger) === null || _a === void 0 ? void 0 : _a.debug("performModeration", logMeta);
            const model = body.model;
            const pool = this.moderationPoolMap.get(model !== null && model !== void 0 ? model : 'text-moderation-latest'); // Get the pool associated with the model
            if (!pool) {
                (_b = this.logger) === null || _b === void 0 ? void 0 : _b.error(`No client pool available for model ${model}`, {}, logMeta);
                throw new Error(`No client pool available for model ${model}`);
            }
            return yield pool.execute((client) => __awaiter(this, void 0, void 0, function* () {
                return yield client.performModeration(body, options);
            }));
        });
    }
    createChatCompletion(createChatCompletionRequest, options, logMeta) {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            (_a = this.logger) === null || _a === void 0 ? void 0 : _a.debug("createChatCompletion", logMeta);
            const model = createChatCompletionRequest.model;
            const pool = this.enrichmentPoolMap.get(model); // Get the pool associated with the model
            if (!pool) {
                (_b = this.logger) === null || _b === void 0 ? void 0 : _b.error(`No client pool available for model ${model}`, {}, logMeta);
                throw new Error(`No client pool available for model ${model}`);
            }
            return yield pool.execute((client) => __awaiter(this, void 0, void 0, function* () {
                return yield client.createChatCompletion(Object.assign(Object.assign({}, createChatCompletionRequest), { stream: false }), options);
            }));
        });
    }
    createStreamingChatCompletion(createChatCompletionRequest, options, logMeta) {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            (_a = this.logger) === null || _a === void 0 ? void 0 : _a.debug("createStreamingChatCompletion", logMeta);
            const model = createChatCompletionRequest.model;
            const pool = this.enrichmentPoolMap.get(model); // Get the pool associated with the model
            if (!pool) {
                (_b = this.logger) === null || _b === void 0 ? void 0 : _b.error(`No client pool available for model ${model}`, {}, logMeta);
                throw new Error(`No client pool available for model ${model}`);
            }
            return yield pool.execute((client) => __awaiter(this, void 0, void 0, function* () {
                return client.createStreamingChatCompletion(Object.assign(Object.assign({}, createChatCompletionRequest), { stream: true }), options);
            }));
        });
    }
}
exports.AIClient = AIClient;
