"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JSONSchemaToTypescriptConverter = void 0;
const utils_1 = require("./utils");
class JSONSchemaToTypescriptConverter {
    constructor(rootSchema, rootName) {
        this.rootName = rootName;
        this.schemaRefStore = {};
        this.typeStore = {};
        this.processSchema(rootSchema, rootName);
    }
    processSchema(schema, name, isInlined = false) {
        if (!isInlined && (schema.$id || (schema.type === "object" && schema.properties))) {
            this.registerSchema(schema, name);
        }
        if (schema.type === "object" && schema.properties) {
            for (const propName in schema.properties) {
                const propSchema = schema.properties[propName];
                if (propSchema) {
                    if (propSchema.type === "object") {
                        const nestedTypeName = this.getNestedTypeName(propSchema, propName);
                        this.processSchema(propSchema, nestedTypeName, !propSchema.$id);
                    }
                    else if (propSchema.type === "array" && propSchema.items) {
                        const nestedSchema = propSchema.items;
                        const nestedTypeName = this.getNestedTypeName(nestedSchema, propName);
                        this.processSchema(nestedSchema, nestedTypeName, !nestedSchema.$id);
                    }
                    else {
                        // console.debug('did not process in schema', propSchema);
                    }
                }
                else {
                    console.debug(`skipping empty property [${propName}] [${JSON.stringify(schema.properties)}]`);
                }
            }
        }
        else if (schema.oneOf) {
            schema.oneOf.forEach((oo, idx) => {
                this.processSchema(oo, this.getNestedTypeName(oo, 'oneOf' + idx), true);
            });
        }
        else if (schema.anyOf) {
            schema.anyOf.forEach((ao, idx) => {
                this.processSchema(ao, this.getNestedTypeName(ao, 'anyOf' + idx), true);
            });
        }
    }
    getTypeStore() {
        return this.typeStore;
    }
    extendFromSchema(sourceSchema) {
        if (sourceSchema.properties) {
            const rootSchema = this.schemaRefStore[this.rootName];
            rootSchema.properties = Object.assign(Object.assign({}, sourceSchema.properties), rootSchema.properties);
            this.processSchema(rootSchema, this.rootName, true);
        }
    }
    registerSchema(schema, id) {
        if (!this.schemaRefStore[id]) {
            this.schemaRefStore[id] = schema;
        }
        if (!this.typeStore[id]) {
            this.typeStore[id] = schema;
        }
    }
    capitalizeFirstLetter(string) {
        return string.charAt(0).toUpperCase() + string.slice(1);
    }
    resolveRef(ref) {
        return this.schemaRefStore[ref.replace('#', '')];
    }
    getNestedTypeName(schema, propName) {
        var _a;
        return (_a = schema.title) !== null && _a !== void 0 ? _a : this.capitalizeFirstLetter(propName);
    }
    // to typescript string stuff
    getTypeString(schema, propName, schemaTrail = [], indent = '') {
        var _a, _b, _c, _d, _e;
        if (schema.anyOf) {
            return schema.anyOf.map((schema) => {
                let out = "";
                if (schema.description && typeof schema.description === 'string') {
                    out = this.genMultiLinecomment(schema.description, out, indent);
                }
                for (const prop in schema.properties) {
                    out += this.stringifyProperties(schema, prop, schemaTrail, indent + '    ');
                }
                return `{\n${out}${indent}}`;
            }).join(' | ');
        }
        else if (schema.oneOf) {
            return schema.oneOf.map((schema) => {
                let out = "";
                if (schema.description && typeof schema.description === 'string') {
                    out = this.genMultiLinecomment(schema.description, out, indent);
                }
                for (const prop in schema.properties) {
                    out += this.stringifyProperties(schema, prop, schemaTrail, indent + '    ');
                }
                return `{\n${out}${indent}}`;
            }).join(' | ');
        }
        else if (schema.$id) {
            //return this.capitalizeFirstLetter(propName);
            return ((_a = schema.title) !== null && _a !== void 0 ? _a : this.capitalizeFirstLetter(schema.$id));
        }
        else if (schema.$ref) {
            const refSchema = this.resolveRef(schema.$ref);
            if (refSchema) {
                return this.capitalizeFirstLetter(propName);
            }
            return 'any';
        }
        else {
            switch (schema.type) {
                case "array":
                    if (schema.items) {
                        if (Array.isArray(schema.items.type)) {
                            const unionTypes = schema.items.type.map((type) => type).join(' | ');
                            return `(${unionTypes})[];`;
                        }
                        else if (schema.items.$ref) {
                            if (schema.items.$ref === '#') {
                                return this.getNestedTypeName([schema, ...schemaTrail].slice(-1)[0], propName) + '[]';
                            }
                            else {
                                return (_b = (0, utils_1.truthy)(this.resolveRef(schema.items.$ref), s => this.getNestedTypeName(s, propName))) !== null && _b !== void 0 ? _b : 'any[]';
                            }
                        }
                        if (schema.items.$id) {
                            return this.getNestedTypeName(schema.items, propName) + '[]';
                        }
                        else {
                            const items = Array.isArray(schema.items) ? schema.items : [schema.items];
                            const types = items.map(item => this.getTypeString(item, propName, [...schemaTrail, schema], indent + '    ')).join(', ');
                            return `${types}[];`;
                        }
                    }
                    return 'any[]';
                case "object":
                    if (schema.properties) {
                        let out = "";
                        for (const prop in schema.properties) {
                            out += this.stringifyProperties(schema, prop, schemaTrail, indent + '    ');
                        }
                        return `{\n${out}${indent}}`;
                    }
                    else {
                        return `${this.capitalizeFirstLetter(propName)};`;
                    }
                default:
                    // primitives..
                    return ((_e = (_d = (_c = (0, utils_1.truthy)(schema.const, _ => JSON.stringify(_))) !== null && _c !== void 0 ? _c : (0, utils_1.truthy)(schema.enum, (en) => `( ${en.map(_ => `'${_}'`).join(' | ')} )`)) !== null && _d !== void 0 ? _d : schema.type) !== null && _e !== void 0 ? _e : 'any');
            }
        }
    }
    genMultiLinecomment(description, out, indent) {
        out += `${indent}/**\n`;
        description.split('\n')
            .forEach((l) => out += `${indent} * ${l}\n`);
        out += `${indent} */\n`;
        return out;
    }
    schemaToString(name, indent = '') {
        const schema = this.typeStore[name];
        let ts = "";
        if (schema.anyOf) {
            ts += `${indent}type ${this.getNestedTypeName(schema, name)} =\n`;
            ts += this.getTypeString(schema, name, [schema], indent + '    ');
            ts += ';\n';
        }
        else if (schema.oneOf) {
            ts += `${indent}type ${this.getNestedTypeName(schema, name)} =\n`;
            ts += this.getTypeString(schema, name, [schema], indent + '    ');
            ts += ';\n';
        }
        else {
            ts += `${indent}interface ${this.getNestedTypeName(schema, name)} {\n`;
            for (const prop in schema.properties) {
                ts += this.stringifyProperties(schema, prop, [schema], indent + '    ');
            }
            ts += `${indent}}\n`;
        }
        // console.debug(`converting [${name}] schema [${JSON.stringify(schema)}] result=[${ts}]}`);
        return ts;
    }
    stringifyProperties(schema, prop, schemaTrail, indent) {
        const propSchema = schema.properties[prop];
        if (typeof propSchema === "boolean")
            return;
        const type = this.getTypeString(propSchema, prop, schemaTrail, indent);
        const description = propSchema.description ? ` // ${propSchema.description}` : '';
        return `${indent}${prop}${propSchema.nullable ? '?' : ''}: ${type} ${description}\n`;
    }
    generateTypescript() {
        let output = "";
        for (const typeName of Object.keys(this.typeStore).reverse()) {
            output += this.schemaToString(typeName);
            output += '\n';
        }
        return output;
    }
}
exports.JSONSchemaToTypescriptConverter = JSONSchemaToTypescriptConverter;
