"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ThrottledTransform = void 0;
const stream_1 = require("stream");
class ThrottledTransform extends stream_1.Transform {
    constructor(options, logger, writeTo) {
        super({ objectMode: true });
        this.flushDebounceTimer = null;
        this.maxTimeoutTimer = null;
        this.cleanMaxTimeout = (scheduleNext = false) => {
            if (this.maxTimeoutTimer) {
                clearTimeout(this.maxTimeoutTimer);
                this.maxTimeoutTimer = null;
            }
            if (scheduleNext) {
                this.maxTimeoutTimer = setTimeout(() => this.flushQueue('maxTimeout', true, 'max timeout reached'), this.options.maxIdleTimeoutMs);
            }
        };
        this.cleanDebounceTimer = () => {
            if (this.flushDebounceTimer) {
                clearTimeout(this.flushDebounceTimer);
                this.flushDebounceTimer = null;
            }
        };
        this.getQueueSize = () => this.queue.length;
        this.logger = logger;
        this.writeTo = writeTo;
        this.queue = [];
        this.options = options;
    }
    _transform(chunk, encoding, callback) {
        var _a, _b;
        (_a = this.logger) === null || _a === void 0 ? void 0 : _a.debug('_transform called', { name: this.options.name });
        this.queue.push(chunk);
        if (this.queue.length >= this.options.windowSize) {
            this.flushQueue('windowSize');
        }
        else {
            if (!this.flushDebounceTimer) {
                (_b = this.logger) === null || _b === void 0 ? void 0 : _b.debug(`_transform setting debounce t=${this.options.flushDebounceTimeMs}`, { name: this.options.name });
                this.flushDebounceTimer = setTimeout(() => this.flushQueue('flushDebounceTime'), this.options.flushDebounceTimeMs);
            }
            this.cleanMaxTimeout(true);
        }
        callback();
    }
    _flush(callback) {
        var _a;
        (_a = this.logger) === null || _a === void 0 ? void 0 : _a.debug('_flush called', { name: this.options.name });
        if (this.queue.length > 0) {
            this.flushQueue('_flush');
        }
        callback();
    }
    flushQueue(reason, forceEnd = false, errorMessage) {
        var _a, _b, _c;
        (_a = this.logger) === null || _a === void 0 ? void 0 : _a.debug(`flushQueue called. reason=[${reason}]`, {
            name: this.options.name,
            forceEnd,
            errorMessage
        });
        this.cleanDebounceTimer();
        if (this.queue.length > 0) {
            // Replace writeToDatabase with your actual function to write to the database
            this.push(this.queue.slice());
            (_b = this.writeTo) === null || _b === void 0 ? void 0 : _b.call(this, this.queue.slice()).catch(e => {
                var _a;
                (_a = this.logger) === null || _a === void 0 ? void 0 : _a.error('error writing to destination', e, { writerName: this.options.name });
                this.emit('error', new Error('error writing to destination'));
            });
            this.queue = [];
        }
        if (forceEnd) {
            if (errorMessage) {
                (_c = this.logger) === null || _c === void 0 ? void 0 : _c.debug('emitting error', { name: this.options.name, forceEnd, errorMessage });
                this.emit('error', new Error(errorMessage));
            }
            else {
                this.emit('end');
            }
        }
    }
    _final(callback) {
        var _a;
        (_a = this.logger) === null || _a === void 0 ? void 0 : _a.debug('final called', { name: this.options.name });
        this.cleanMaxTimeout();
        this.cleanDebounceTimer();
        super._final(callback);
    }
}
exports.ThrottledTransform = ThrottledTransform;
